# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Use the kiosk-frontend API to create and monitor image processing jobs"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import logging
import logging.handlers
import os
import sys

from twisted.internet import reactor

from kiosk_client import manager
from kiosk_client import settings


def valid_filepath(parser, arg):
    """File validation for argparsing.

    https://stackoverflow.com/a/11541450
    """
    if not os.path.exists(arg):
        # Argparse uses the ArgumentTypeError to give a rejection message like:
        # error: argument input: x does not exist
        raise argparse.ArgumentTypeError('{0} does not exist'.format(arg))
    return arg


def get_arg_parser():
    parser = argparse.ArgumentParser(
        prog='kiosk_client',
        description='The Kiosk-Client is a Command Line Interface (CLI) for '
                    'interacting with the DeepCell Kiosk.'
    )

    # Job definition
    parser.add_argument('file', type=str, metavar="FILE",
                        help='File to process in many duplicated jobs. '
                             '(Must exist in the cloud storage bucket if '
                             'using benchmark mode.)')

    parser.add_argument('--benchmark', action='store_true',
                        help='Benchmarking mode. Manage COUNT simulated jobs. '
                             'The FILE must exist in the STORAGE_BUCKET.')

    parser.add_argument('-j', '--job-type', type=str, required=True,
                        help='Type of job (name of Redis work queue).')

    parser.add_argument('-t', '--host', type=str, required=True,
                        help='IP or FQDN of the DeepCell Kiosk API.')

    parser.add_argument('-m', '--model', type=str,
                        default=settings.MODEL,
                        help='Name and version of model hosted by TensorFlow '
                             'Serving. Overrides the default model defined by '
                             'the JOB_TYPE.')

    parser.add_argument('-b', '--storage-bucket', type=str,
                        default=settings.STORAGE_BUCKET,
                        help='Cloud storage bucket (e.g. gs://storage-bucket).'
                             'Only required if using `--upload-results`.')

    parser.add_argument('-c', '--count', default=1, type=int,
                        help='Number of times to process the given file. '
                             'Only used in `benchmark` mode.')

    parser.add_argument('--pre', '--preprocess', type=str,
                        default=settings.PREPROCESS,
                        help='Preprocessing function to use before model '
                             'prediction. Overrides default preprocessing '
                             'function for the JOB_TYPE.')

    parser.add_argument('--post', '--postprocess', type=str,
                        default=settings.POSTPROCESS,
                        help='Postprocessing function to use after model '
                             'prediction. Overrides default postprocessing '
                             'function for the JOB_TYPE.')

    parser.add_argument('-s', '--scale', type=str,
                        default=settings.SCALE,
                        help='Scale of the data. Data will be scaled up or '
                             'for the best model compatibility.')

    parser.add_argument('-l', '--label', type=str,
                        default=settings.LABEL, choices=['', '0', '1', '2'],
                        help='Data type (e.g. nuclear, cytoplasmic, etc.).')

    parser.add_argument('-U', '--upload', action='store_true',
                        help='If provided, uploads the file before creating '
                             'a new job. '
                             '(Only applicable in `benchmark` mode.)')

    parser.add_argument('--upload-results', action='store_true',
                        help='Upload the final output file to the bucket.')

    parser.add_argument('--no-download-results', action='store_true',
                        help='Upload the final output file to the bucket.')

    parser.add_argument('--calculate-cost', action='store_true',
                        help='Use the Grafana API to calculate the cost of '
                             'the job.')

    # Timing / interval settings
    parser.add_argument('--start-delay', type=float,
                        default=settings.START_DELAY,
                        help='Time between each job creation '
                             '(0.5s is a typical file upload time).')

    parser.add_argument('--update-interval', type=float,
                        default=settings.UPDATE_INTERVAL,
                        help='Seconds between each job status refresh.')

    parser.add_argument('--refresh-rate', type=float,
                        default=settings.MANAGER_REFRESH_RATE,
                        help='Seconds between each manager status check.')

    parser.add_argument('-x', '--expire-time', type=float,
                        default=settings.EXPIRE_TIME,
                        help='Finished jobs expire after this many seconds.')

    # Logging options
    parser.add_argument('-L', '--log-level', default=settings.LOG_LEVEL,
                        choices=('DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'),
                        help='Only log the given level and above.')

    # optional arguments
    parser.add_argument('--upload-prefix', type=str,
                        default=settings.UPLOAD_PREFIX,
                        help='Maximum number of connections to the Kiosk.')

    return parser


def initialize_logger(log_level):
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    log_level = getattr(logging, log_level)

    formatter = logging.Formatter(fmt=settings.LOG_FORMAT)

    console = logging.StreamHandler(stream=sys.stdout)
    console.setFormatter(formatter)
    console.setLevel(log_level)
    logger.addHandler(console)

    fh = logging.handlers.RotatingFileHandler(
        filename=settings.LOG_FILE,
        maxBytes=10000000,
        backupCount=1)
    fh.setFormatter(formatter)
    fh.setLevel(log_level)
    logger.addHandler(fh)

    logging.getLogger('PIL').setLevel(logging.INFO)


if __name__ == '__main__':
    args = get_arg_parser().parse_args()

    if settings.LOG_ENABLED:
        initialize_logger(log_level=args.log_level)

    if args.scale:  # optional, but if provided should be a float
        try:
            args.scale = float(args.scale)
        except ValueError:
            raise argparse.ArgumentTypeError(
                '{0} is not a float'.format(args.scale))

    mgr_kwargs = {
        'host': args.host,
        'model': args.model,
        'job_type': args.job_type,
        'update_interval': args.update_interval,
        'start_delay': args.start_delay,
        'refresh_rate': args.refresh_rate,
        'postprocess': args.post,
        'preprocess': args.pre,
        'upload_prefix': args.upload_prefix,
        'expire_time': args.expire_time,
        'data_scale': args.scale,
        'data_label': args.label,
        'storage_bucket': args.storage_bucket,
        'upload_results': args.upload_results,
        'calculate_cost': args.calculate_cost,
        'download_results': not args.no_download_results,
    }

    if not os.path.exists(args.file) and not args.benchmark and args.upload:
        raise FileNotFoundError('%s could not be found.' % args.file)

    if args.benchmark:
        mgr = manager.BenchmarkingJobManager(**mgr_kwargs)
        mgr.run(filepath=args.file, count=args.count, upload=args.upload)

    else:
        mgr = manager.BatchProcessingJobManager(**mgr_kwargs)
        mgr.run(filepath=args.file)

    reactor.run()  # pylint: disable=E1101



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utility files for batch file processing"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from PIL import Image

from twisted.internet import reactor
from twisted.internet.task import deferLater


def get_download_path():
    """Returns the default downloads path for linux or windows.
    https://stackoverflow.com/a/48706260
    """
    if os.name == 'nt':
        # pylint: disable=E0401,C0415
        import winreg
        k = r'SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
        downloads_guid = '{374DE290-123F-4565-9164-39C4925E467B}'
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, k) as key:
            location = winreg.QueryValueEx(key, downloads_guid)[0]
        return location
    location = os.path.join(os.path.expanduser('~'), 'Downloads')
    while not os.path.exists(location):
        location = os.path.abspath(os.path.join(location, '..'))
    return location


def strip_bucket_prefix(prefix):
    """Remove any leading or trailing "/" characters."""
    return '/'.join(x for x in prefix.split('/') if x)


def sleep(seconds):
    """Simple helper to delay asynchronously for some number of seconds."""
    return deferLater(reactor, seconds, lambda: None)


def is_image_file(filepath):
    """Returns True if the file is an image file, otherwise False"""
    try:
        with Image.open(filepath) as im:
            im.verify()
        return True
    except:  # pylint: disable=bare-except
        return False


def iter_image_files(path, include_archives=True):
    archive_extensions = {'.zip'}
    if os.path.isfile(path):
        _, ext = os.path.splitext(path.lower())
        if ext in archive_extensions and include_archives:
            yield path
        elif is_image_file(path):
            yield path

    for (dirpath, _, filenames) in os.walk(path):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            # process all zip images
            _, ext = os.path.splitext(filepath.lower())
            if ext in archive_extensions and include_archives:
                yield filepath

            # process all images
            elif is_image_file(filepath):
                yield filepath



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Manager class used to create and manage jobs"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
import json
import logging
import os
import shutil
import timeit
import uuid
import zipfile

import requests
from google.cloud import storage as google_storage
from twisted.internet import defer, reactor
from twisted.web.client import HTTPConnectionPool

from kiosk_client.job import Job
from kiosk_client.utils import iter_image_files
from kiosk_client.utils import sleep
from kiosk_client.utils import strip_bucket_prefix
from kiosk_client import settings

from kiosk_client.cost import CostGetter


class JobManager(object):
    """Manages many DeepCell Kiosk jobs.

    Args:
        host (str): public IP address of the DeepCell Kiosk cluster.
        job_type (str): DeepCell Kiosk job type (e.g. "segmentation").
        upload_prefix (str): upload all files to this folder in the bucket.
        refresh_rate (int): seconds between each manager status check.
        update_interval (int): seconds between each job status refresh.
        expire_time (int): seconds until finished jobs are expired.
        start_delay (int): delay between each job, in seconds.
    """

    def __init__(self, host, job_type, **kwargs):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.created_at = timeit.default_timer()
        self.all_jobs = []

        self.host = self._get_host(host)
        self.job_type = job_type

        model = kwargs.get('model', '')
        if model:
            try:
                model_name, model_version = str(model).split(':')
                model_version = int(model_version)
            except Exception as err:
                self.logger.error('Invalid model name, must be of the form '
                                  '"ModelName:Version", for example "model:0".')
                raise err
        else:
            model_name, model_version = '', ''

        self.model_name = model_name
        self.model_version = model_version

        data_scale = str(kwargs.get('data_scale', ''))
        if data_scale:
            try:
                data_scale = float(data_scale)
            except ValueError:
                raise ValueError('data_scale must be a number.')
        self.data_scale = data_scale

        data_label = str(kwargs.get('data_label', ''))
        if data_label:
            try:
                data_label = int(data_label)
            except ValueError:
                raise ValueError('data_label must be an integer.')
        self.data_label = data_label

        self.preprocess = kwargs.get('preprocess', '')
        self.postprocess = kwargs.get('postprocess', '')
        self.upload_prefix = kwargs.get('upload_prefix', 'uploads')
        self.upload_prefix = strip_bucket_prefix(self.upload_prefix)
        self.refresh_rate = int(kwargs.get('refresh_rate', 10))
        self.update_interval = kwargs.get('update_interval', 10)
        self.expire_time = kwargs.get('expire_time', 3600)
        self.start_delay = kwargs.get('start_delay', 0.1)
        self.bucket = kwargs.get('storage_bucket')
        self.upload_results = kwargs.get('upload_results', False)
        self.download_results = kwargs.get('download_results', True)
        self.calculate_cost = kwargs.get('calculate_cost', False)

        # initializing cost estimation workflow
        self.cost_getter = CostGetter()

        self.sleep = sleep  # allow monkey-patch

        # twisted configuration
        self.pool = HTTPConnectionPool(reactor, persistent=True)
        self.pool.maxPersistentPerHost = settings.CONCURRENT_REQUESTS_PER_HOST
        self.pool.retryAutomatically = False

    def _get_host(self, host):
        """Send a GET request to the provided host. Check for redirects.

        Twisted does not allow POST requests to follow redirects. Use requests
        to send a single GET request to the host and follow any redirects.

        Args:
            host (str): The user-provided hostname.

        Returns:
            str: The hostname after all redirects.
        """
        host = str(host).lower()
        if not any(host.startswith(x) for x in ('http://', 'https://')):
            host = 'http://{}'.format(host)

        try:
            url = str(requests.get(host).url)
            url = url[:-1] if url.endswith('/') else url
            return url
        except:
            raise RuntimeError('Could not connect to host: %s' % host)

    def upload_file(self, filepath, acl='publicRead',
                    hash_filename=True, prefix=None):
        prefix = self.upload_prefix if prefix is None else prefix
        start = timeit.default_timer()
        storage_client = google_storage.Client()

        self.logger.debug('Uploading %s.', filepath)
        if hash_filename:
            _, ext = os.path.splitext(filepath)
            dest = '{}{}'.format(uuid.uuid4().hex, ext)
        else:
            dest = os.path.basename(filepath)

        bucket = storage_client.get_bucket(self.bucket)
        blob = bucket.blob(os.path.join(prefix, dest))
        blob.upload_from_filename(filepath, predefined_acl=acl)
        self.logger.debug('Uploaded %s to %s in %s seconds.',
                          filepath, dest, timeit.default_timer() - start)
        return dest

    def make_job(self, filepath):
        return Job(filepath=filepath,
                   host=self.host,
                   model_name=self.model_name,
                   model_version=self.model_version,
                   job_type=self.job_type,
                   data_scale=self.data_scale,
                   data_label=self.data_label,
                   postprocess=self.postprocess,
                   upload_prefix=self.upload_prefix,
                   update_interval=self.update_interval,
                   download_results=self.download_results,
                   expire_time=self.expire_time,
                   pool=self.pool)

    def get_completed_job_count(self):
        created, complete, failed, expired = 0, 0, 0, 0

        statuses = {}

        for j in self.all_jobs:
            expired += int(j.is_expired)  # true mark of being done
            complete += int(j.is_summarized)
            created += int(j.job_id is not None)

            if j.status is not None:
                if j.status not in statuses:
                    statuses[j.status] = 1
                else:
                    statuses[j.status] += 1

            if j.failed:
                j.restart(delay=self.start_delay * failed)

            # # TODO: patched! "done" jobs can get stranded before summarization
            # if j.status == 'done' and not j.is_summarized:
            #     j.summarize()
            #
            # # TODO: patched! sometimes jobs don't get expired?
            # elif j.status == 'done' and j.is_summarized and not j.is_expired:
            #     j.expire()

        self.logger.info('%s created; %s finished; %s summarized; '
                         '%s; %s jobs total', created, expired, complete,
                         '; '.join('%s %s' % (v, k)
                                   for k, v in statuses.items()),
                         len(self.all_jobs))

        if len(self.all_jobs) - expired <= 25:
            for j in self.all_jobs:
                if not j.is_expired:
                    self.logger.info('Waiting on key `%s` with status %s',
                                     j.job_id, j.status)

        return expired

    @defer.inlineCallbacks
    def _stop(self):
        yield reactor.stop()  # pylint: disable=no-member

    @defer.inlineCallbacks
    def check_job_status(self):
        complete = -1  # initialize comparison value

        while complete != len(self.all_jobs):
            yield self.sleep(self.refresh_rate)

            complete = self.get_completed_job_count()  # synchronous

        self.summarize()  # synchronous

        yield self._stop()

    def summarize(self):
        time_elapsed = timeit.default_timer() - self.created_at
        self.logger.info('Finished %s jobs in %s seconds.',
                         len(self.all_jobs), time_elapsed)

        # add cost and timing data to json output
        cpu_cost, gpu_cost, total_cost = '', '', ''
        if self.calculate_cost:
            try:
                cpu_cost, gpu_cost, total_cost = self.cost_getter.finish()
            except Exception as err:  # pylint: disable=broad-except
                self.logger.error('Encountered %s while getting cost data: %s',
                                  type(err).__name__, err)

        jsondata = {
            'cpu_node_cost': cpu_cost,
            'gpu_node_cost': gpu_cost,
            'total_node_and_networking_costs': total_cost,
            'start_delay': self.start_delay,
            'num_jobs': len(self.all_jobs),
            'time_elapsed': time_elapsed,
            'job_data': [j.json() for j in self.all_jobs]
        }

        output_filepath = '{}{}jobs_{}delay_{}.json'.format(
            '{}gpu_'.format(settings.NUM_GPUS) if settings.NUM_GPUS else '',
            len(self.all_jobs), self.start_delay, uuid.uuid4().hex)
        output_filepath = os.path.join(settings.OUTPUT_DIR, output_filepath)

        with open(output_filepath, 'w') as jsonfile:
            json.dump(jsondata, jsonfile, indent=4)
            self.logger.info('Wrote job data as JSON to %s.', output_filepath)

        if self.upload_results:
            try:
                _ = self.upload_file(output_filepath,
                                     hash_filename=False,
                                     prefix='output')
            except Exception as err:  # pylint: disable=broad-except
                self.logger.error(err)
                self.logger.error('Could not upload output file to bucket. '
                                  'Copy this file from the docker container to '
                                  'keep the data.')

    def run(self, *args, **kwargs):
        raise NotImplementedError


class BenchmarkingJobManager(JobManager):
    # pylint: disable=arguments-differ

    @defer.inlineCallbacks
    def run(self, filepath, count, upload=False):
        self.logger.info('Benchmarking %s jobs of file `%s`', count, filepath)

        for i in range(count):

            job = self.make_job(filepath)

            self.all_jobs.append(job)

            # stagger the delay seconds; if upload it will be staggered already
            job.start(delay=self.start_delay * i * int(not upload),
                      upload=upload)

            yield self.sleep(self.start_delay * upload)

            if upload:
                self.get_completed_job_count()  # log during uploading

        yield self.check_job_status()


class BatchProcessingJobManager(JobManager):
    # pylint: disable=arguments-differ

    @defer.inlineCallbacks
    def run(self, filepath):
        self.logger.info('Benchmarking all image/zip files in `%s`', filepath)

        for i, f in enumerate(iter_image_files(filepath)):
            job = self.make_job(f)
            self.all_jobs.append(job)
            # stagger the delay seconds
            job.start(delay=self.start_delay * i, upload=True)

        yield self.check_job_status()



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for utility functions"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import random
import zipfile

from PIL import Image
import pytest

from kiosk_client import utils


class TestUtils(object):

    def test_get_download_path(self):
        download_dir = utils.get_download_path()
        assert os.path.isdir(download_dir)

    def test_is_image_file(self, tmpdir):
        # Test valid image
        tmpdir = str(tmpdir)
        valid_image = os.path.join(tmpdir, 'image.png')
        img = Image.new('RGB', (800, 1280), (255, 255, 255))
        img.save(valid_image, 'PNG')
        assert utils.is_image_file(valid_image)

        # Test invalid image
        invalid_image = os.path.join(tmpdir, 'bad_image.png')
        with open(invalid_image, 'w') as f:
            f.write('line1')
        assert not utils.is_image_file(invalid_image)

        # Test image file does not exist
        missing_image = os.path.join(tmpdir, 'nofile.png')
        assert not utils.is_image_file(missing_image)

    def test_iter_image_files(self, tmpdir):
        # test image files
        tmpdir = str(tmpdir)
        num = random.randint(1, 5)
        imagename = lambda x: 'image%s.png' % x

        valid_images = []
        for i in range(num):
            valid_image = os.path.join(tmpdir, imagename(i))
            img = Image.new('RGB', (800, 1280), (255, 255, 255))
            img.save(valid_image, 'PNG')
            valid_images.append(valid_image)

        # only image files exist
        results = utils.iter_image_files(tmpdir, include_archives=True)
        assert set(list(results)) == set(valid_images)

        # create a new zip file of all the valid images
        zippath = os.path.join(tmpdir, 'test.zip')
        z = zipfile.ZipFile(zippath, 'w', zipfile.ZIP_DEFLATED)
        z.close()

        # turn off include_archives, should not see the zip file
        results = utils.iter_image_files(tmpdir, include_archives=False)
        assert set(list(results)) == set(valid_images)

        # with include_archives, zipfile should be in results
        results = utils.iter_image_files(tmpdir, include_archives=True)
        assert set(list(results)) == set(valid_images).union({zippath})

        # test single file paths
        results = utils.iter_image_files(zippath, include_archives=True)
        assert set(list(results)) == set((zippath,))
        results = utils.iter_image_files(zippath, include_archives=False)
        assert set(list(results)) == set()
        results = utils.iter_image_files(valid_images[0])
        assert set(list(results)) == set((valid_images[0],))

    def test_strip_bucket_prefix(self):
        names = [
            'uploads',
            'uploads/with/subdir',
        ]
        fmt_strings = [
            '/{}',  # leading "/"
            '{}/',  # trailing "/"
            '/{}/',  # both leading and trailing "/"
            '{}',  # nothing
        ]
        for name in names:
            for fmt_string in fmt_strings:
                prefix = fmt_string.format(name)
                assert name == utils.strip_bucket_prefix(prefix)

        name = '/duplicate/leading/and/trailing/'
        stripped = utils.strip_bucket_prefix(name)
        assert name[1:-1] == stripped



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Class to estimate costs using the grafana API"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import time
import urllib

import requests

from kiosk_client import settings


# defining Google Cloud prices

# computed networking costs for 1,000,000 image run:
NETWORKING_COSTS = 7.00

# current, as of 6/17/19
COST_TABLE = {
    'n1-standard-1': {'ondemand': 0.0475, 'preemptible': 0.0100},
    'n1-standard-2': {'ondemand': 0.0950, 'preemptible': 0.0200},
    'n1-standard-4': {'ondemand': 0.1900, 'preemptible': 0.0400},
    'n1-standard-8': {'ondemand': 0.3800, 'preemptible': 0.0800},
    'n1-standard-16': {'ondemand': 0.7600, 'preemptible': 0.1600},
    'n1-standard-32': {'ondemand': 1.5200, 'preemptible': 0.3200},
    'n1-standard-64': {'ondemand': 3.0400, 'preemptible': 0.6400},
    'n1-standard-96': {'ondemand': 4.5600, 'preemptible': 0.9600},
    'n1-highmem-2': {'ondemand': 0.1184, 'preemptible': 0.0250},
    'n1-highmem-4': {'ondemand': 0.2368, 'preemptible': 0.0500},
    'n1-highmem-8': {'ondemand': 0.4736, 'preemptible': 0.1000},
    'n1-highmem-16': {'ondemand': 0.9472, 'preemptible': 0.2000},
    'n1-highmem-32': {'ondemand': 1.8944, 'preemptible': 0.4000},
    'n1-highmem-64': {'ondemand': 3.7888, 'preemptible': 0.8000},
    'n1-highmem-96': {'ondemand': 5.6832, 'preemptible': 1.2000},
    'n1-highcpu-2': {'ondemand': 0.0709, 'preemptible': 0.0150},
    'n1-highcpu-4': {'ondemand': 0.1418, 'preemptible': 0.0300},
    'n1-highcpu-8': {'ondemand': 0.2836, 'preemptible': 0.0600},
    'n1-highcpu-16': {'ondemand': 0.5672, 'preemptible': 0.1200},
    'n1-highcpu-32': {'ondemand': 1.1344, 'preemptible': 0.2400},
    'n1-highcpu-64': {'ondemand': 2.2688, 'preemptible': 0.4800},
    'n1-highcpu-96': {'ondemand': 3.4020, 'preemptible': 0.7200},
    'n1-ultramem-40': {'ondemand': 6.3039, 'preemptible': 1.3311},
    'n1-ultramem-80': {'ondemand': 12.6078, 'preemptible': 2.6622},
    'n1-ultramem-160': {'ondemand': 25.2156, 'preemptible': 5.3244},
    'n1-megamem-96': {'ondemand': 10.6740, 'preemptible': 2.2600},
}

GPU_TABLE = {
    'nvidia-tesla-t4': {'ondemand': 0.95, 'preemptible': 0.29},
    'nvidia-tesla-p4': {'ondemand': 0.60, 'preemptible': 0.216},
    'nvidia-tesla-v100': {'ondemand': 2.48, 'preemptible': 0.74},
    'nvidia-tesla-p100': {'ondemand': 1.46, 'preemptible': 0.43},
    'nvidia-tesla-k80': {'ondemand': 0.45, 'preemptible': 0.135},
}


class CostGetter(object):

    def __init__(self,
                 benchmarking_start_time=None,
                 benchmarking_end_time=None,
                 **kwargs):
        # validate benchmarking_start_time
        if benchmarking_start_time is None:
            benchmarking_start_time = self.get_time()

        try:
            benchmarking_start_time = int(benchmarking_start_time)
            assert benchmarking_start_time <= int(time.time())
        except AssertionError:
            raise ValueError('You need to provide either an integer '
                             'string (e.g., "12345") or a decimal or '
                             'integer number (e.g., 123.45) for '
                             'benchmarking_start_time.')

        if benchmarking_end_time is not None:
            try:
                benchmarking_end_time = int(benchmarking_end_time)
                assert benchmarking_start_time <= benchmarking_end_time
            except AssertionError:
                raise ValueError('You need to provide either an integer '
                                 'string (e.g., "12345") or a decimal or '
                                 'integer number (e.g., 123.45) for '
                                 'benchmarking_end_time.')

        self.benchmarking_start_time = benchmarking_start_time
        self.benchmarking_end_time = benchmarking_end_time

        # initialize other necessary variables
        self.min_step = 15
        self.cost_table = kwargs.get('cost_table', COST_TABLE)
        self.gpu_table = kwargs.get('gpu_table', GPU_TABLE)
        self.networking_costs = kwargs.get('networking_costs', NETWORKING_COSTS)

        self.grafana_user = settings.GRAFANA_USER
        self.grafana_password = settings.GRAFANA_PASSWORD
        self.grafana_host = settings.GRAFANA_HOST
        self.logger = logging.getLogger(str(self.__class__.__name__))

    @classmethod
    def get_time(cls):
        """Get current time in epoch seconds."""
        # Meant to be called at the beginning and end of a benchmarking run
        # to establish the beginning or end of cost accrual.
        return int(time.time())

    def get_query_data(self, query, step=None):
        """Return a payload of the given query for the Grafana API"""
        step = self.min_step if step is None else step
        return {
            'query': query,
            'start': self.benchmarking_start_time,
            'end': self.benchmarking_end_time,
            'step': step,
        }

    def get_url(self, data):
        """Return a formatted URL for the Grafana API"""
        # check python2 vs python3
        if hasattr(urllib, 'parse'):
            url_encode = urllib.parse.urlencode  # pylint: disable=E1101
        else:
            url_encode = urllib.urlencode  # pylint: disable=E1101
        return 'http://{user}:{passwd}@{host}{route}?{querystring}'.format(
            user=self.grafana_user,
            passwd=self.grafana_password,
            host=self.grafana_host,
            route='/api/datasources/proxy/1/api/v1/query_range',
            querystring=url_encode(data))

    def send_grafana_api_request(self, query, step=None):
        """Send a HTTP GET request with the data url encoded"""
        # initialize retry loop values
        status_code = None
        errortext = None
        is_first_req = True

        # error text found in requests that are too large. must increase step.
        retryable_errortext = 'exceeded maximum resolution'

        reqdata = self.get_query_data(query, step)

        status_is_400 = lambda x: x is None or x == 400
        retryable_error = lambda x: x is None or retryable_errortext in str(x)

        # if there are too many datapoints, a 400 error code is returned.
        # inspect the error text to confirm.
        while status_is_400(status_code) and retryable_error(errortext):
            # don't spam the API
            time.sleep(0.5 * int(not is_first_req))

            reqdata['step'] += self.min_step * int(not is_first_req)
            url = self.get_url(reqdata)

            response = requests.get(url)
            status_code = response.status_code

            jsondata = response.json()

            if status_code != 200:
                is_first_req = False  # starting to retry
                errortext = jsondata.get('error', '')
                self.logger.warning('%s request failed due to error: %s',
                                    query, errortext)

        return jsondata

    def finish(self):
        # This is the wrapper function for all the functionality
        # that will executed immediately once benchmarking is finished.
        if not self.benchmarking_end_time:
            self.benchmarking_end_time = self.get_time()

        creation_data = self.send_grafana_api_request('kube_node_created')
        label_data = self.send_grafana_api_request('kube_node_labels')

        parsed_creation_data = self.parse_create_response(creation_data)
        parsed_label_data = self.parse_label_response(label_data)

        # merge the dictionaries
        node_data = parsed_creation_data.copy()
        for node_name in parsed_label_data:
            if node_name in node_data:
                for k in parsed_label_data[node_name]:
                    node_data[node_name][k] = parsed_label_data[node_name][k]

        (cpu_node_costs, gpu_node_costs, total_node_costs) = \
            self.compute_costs(node_data)
        total_costs = total_node_costs + self.networking_costs
        return str(cpu_node_costs), str(gpu_node_costs), str(total_costs)

    def parse_create_response(self, response):
        node_info = {}
        # parse node liveness data
        for time_series in response['data']['result']:
            # get node name and create dictionary entry
            node_name = time_series['metric']['node']
            node_info[node_name] = {}
            # get node lifetime
            # Was there only one creation event?
            first_event = time_series['values'][0]
            last_event = time_series['values'][-1]

            if first_event[-1] == last_event[-1]:
                # they're the same! easy to calculate
                created_at = int(last_event[-1])
                # only count costs during the benchmarking window.
                created_at = max(created_at, self.benchmarking_start_time)
                node_info[node_name]['lifetime'] = last_event[0] - created_at
                continue

            # there was more than one creation event :(
            # loop over list backward to find the final event for each creation
            lifetime = 0
            curr_label = None
            for i in range(len(time_series['values']) - 1, 0, -1):
                ts, created_at = time_series['values'][i]
                created_at = int(created_at)
                if created_at != curr_label:  # a new label.
                    curr_label = created_at
                    created_at = max(created_at, self.benchmarking_start_time)
                    lifetime += ts - created_at

            node_info[node_name]['lifetime'] = lifetime
        return node_info

    def parse_label_response(self, response):
        node_info = {}
        # parse node label data
        for label_set in response['data']['result']:
            metric = label_set['metric']

            node_name = metric.get('label_kubernetes_io_hostname')
            node_info[node_name] = {}

            instance_type = metric['label_beta_kubernetes_io_instance_type']
            preemptible = 'label_cloud_google_com_gke_preemptible' in metric
            gpu = metric.get('label_cloud_google_com_gke_accelerator')

            node_info[node_name]['instance_type'] = instance_type
            node_info[node_name]['preemptible'] = preemptible
            node_info[node_name]['gpu'] = gpu

        return node_info

    def compute_costs(self, node_data):
        """Get cost for all nodes"""
        gpu_node_costs = 0
        cpu_node_costs = 0
        total_node_costs = 0
        for _, node_dict in node_data.items():
            node_hourly_cost = self.compute_hourly_cost(node_dict)
            node_cost = node_hourly_cost * (node_dict['lifetime'] / 60 / 60)
            if not node_dict['gpu']:
                cpu_node_costs = cpu_node_costs + node_cost
            else:
                gpu_node_costs = gpu_node_costs + node_cost
            total_node_costs = total_node_costs + node_cost
        return cpu_node_costs, gpu_node_costs, total_node_costs

    def compute_hourly_cost(self, node_data):
        """Get the hourly cost of a given node"""
        instance_type = node_data['instance_type']
        gpu = node_data['gpu']
        preemptible = node_data['preemptible']

        # which types of nodes
        key = 'ondemand' if not preemptible else 'preemptible'

        instance_cost = self.cost_table[instance_type][key]

        gpu_cost = 0 if not gpu else self.gpu_table[gpu][key]

        hourly_cost = instance_cost + gpu_cost
        return hourly_cost



# Copyright 2016-2018 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Constants and environment variables"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import errno
import os

from decouple import config


NUM_GPUS = config('NUM_GPUS', cast=int, default=0)

# Google credentials
STORAGE_BUCKET = config('STORAGE_BUCKET', default='')

# Batch API Host (IP Address or FQDN)
HOST = config('API_HOST', cast=str, default='')
if not any(HOST.lower().startswith(x) for x in ('http://', 'https://')):
    HOST = 'http://{}'.format(HOST)

# Grafana resources for cost estimation
GRAFANA_HOST = config('GRAFANA_HOST', default='prometheus-operator-grafana')
GRAFANA_USER = config('GRAFANA_USER', default='admin')
GRAFANA_PASSWORD = config('GRAFANA_PASSWORD', default='prom-operator')

# TensorFlow Servable
MODEL = config('MODEL', default='')

# Job Type
JOB_TYPE = config('JOB_TYPE', default='segmentation')
SCALE = config('SCALE', default='')  # detect scale automatically if empty
LABEL = config('LABEL', default='')   # detect data type automatically if empty

# Pre- and Post-Processing functions
PREPROCESS = config('PREPROCESS', default='')
POSTPROCESS = config('POSTPROCESS', default='')

# How frequently Jobs update their statuses
UPDATE_INTERVAL = config('UPDATE_INTERVAL', default=10, cast=float)

# Time to wait between starting jobs (for staggering redis entries)
START_DELAY = config('START_DELAY', default=0.05, cast=float)

# Time interval between Manager status checks
MANAGER_REFRESH_RATE = config('MANAGER_REFRESH_RATE', default=10, cast=float)

# Time in seconds to expire the completed jobs.
EXPIRE_TIME = config('EXPIRE_TIME', default=3600, cast=int)

# Name of upload folder in storage bucket.
UPLOAD_PREFIX = config('UPLOAD_PREFIX', default='uploads', cast=str)

# HTTP Settings
CONCURRENT_REQUESTS_PER_HOST = config('CONCURRENT_REQUESTS_PER_HOST',
                                      default=64, cast=int)

# Application directories
ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DOWNLOAD_DIR = os.path.join(ROOT_DIR, 'download')
OUTPUT_DIR = os.path.join(ROOT_DIR, 'output')
LOG_DIR = os.path.join(ROOT_DIR, 'logs')

# Log settings
LOG_ENABLED = config('LOG_ENABLED', default=True, cast=bool)
LOG_FORMAT = '[%(asctime)s]:[%(levelname)s]:[%(name)s]: %(message)s'
LOG_LEVEL = config('LOG_LEVEL', cast=str, default='DEBUG')
LOG_FILE = config('LOG_FILE', default='benchmark.log')
LOG_FILE = os.path.join(LOG_DIR, LOG_FILE)

# Overwrite directories with environment variabls
DOWNLOAD_DIR = config('DOWNLOAD_DIR', default=DOWNLOAD_DIR)
OUTPUT_DIR = config('OUTPUT_DIR', default=OUTPUT_DIR)
LOG_DIR = config('LOG_DIR', default=LOG_DIR)

for d in (DOWNLOAD_DIR, OUTPUT_DIR, LOG_DIR):
    try:
        os.makedirs(d)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for the file cost_estimator.py"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random
import time

import pytest
import requests.exceptions

from kiosk_client import cost


class FakeCreationData:

    _status_code = 400

    @property
    def status_code(self):
        code = self._status_code
        if code == 400:
            self._status_code = 200

    @staticmethod
    def json(created_at=None, lifetime=10):
        if created_at is None:
            created_at = int(time.time())
        return {
            'data': {
                'result': [
                    {
                        'metric': {
                            '__name__': 'kube_node_created',
                            'endpoint': 'http',
                            'instance': '10.48.5.8:8080',
                            'job': 'kube-state-metrics',
                            'namespace': 'test',
                            'node': 'test_node_1',
                            'pod': 'prometheus-operator-kube-state-metrics-1',
                            'service': 'prometheus-operator-kube-state-metrics'
                        },
                        'values': [
                            [
                                created_at,
                                str(created_at)
                            ],
                            [
                                created_at + lifetime,
                                str(created_at)
                            ]
                        ]
                    },
                    {
                        'metric': {
                            '__name__': 'kube_node_created',
                            'endpoint': 'http',
                            'instance': '10.48.5.9:8080',
                            'job': 'kube-state-metrics',
                            'namespace': 'test',
                            'node': 'test_node_2',
                            'pod': 'prometheus-operator-kube-state-metrics-1',
                            'service': 'prometheus-operator-kube-state-metrics'
                        },
                        'values': [
                            [
                                created_at,
                                str(created_at)
                            ],
                            [
                                created_at + lifetime - lifetime // 2,
                                str(created_at)
                            ],
                            [
                                created_at + lifetime - lifetime // 2,
                                str(created_at + lifetime - lifetime // 2)
                            ],
                            [
                                created_at + lifetime - 1,
                                str(created_at + lifetime - lifetime // 2)
                            ],
                            [
                                created_at + lifetime,
                                str(created_at + lifetime - lifetime // 2)
                            ]
                        ]
                    },
                ],
                'resultType': 'matrix',
            },
            'status': 'success',
        }


class FakeLabelData:

    _status_code = 400

    @property
    def status_code(self):
        code = self._status_code
        if code == 400:
            self._status_code = 200

    @staticmethod
    def json(created_at=None, lifetime=10):
        if created_at is None:
            created_at = int(time.time())
        return {
            'data': {
                'result': [
                    {
                        'metric': {
                            '__name__': 'kube_node_labels',
                            'endpoint': 'http',
                            'instance': '10.48.5.8:8080',
                            'job': 'kube-state-metrics',
                            'label_beta_kubernetes_io_arch': 'amd64',
                            'label_beta_kubernetes_io_fluentd_ds_ready':
                                'true',
                            'label_beta_kubernetes_io_instance_type':
                                'n1-highmem-2',
                            'label_beta_kubernetes_io_os': 'linux',
                            'label_cloud_google_com_gke_accelerator':
                                'nvidia-tesla-v100',
                            'label_cloud_google_com_gke_nodepool':
                                'prediction-gpu',
                            'label_cloud_google_com_gke_os_distribution':
                                'cos',
                            'label_cloud_google_com_gke_preemptible': 'true',
                            'label_failure_domain_beta_kubernetes_io_region':
                                'us-west1',
                            'label_failure_domain_beta_kubernetes_io_zone':
                                'us-west1-a',
                            'label_kubernetes_io_hostname': 'test_node_1',
                            'namespace': 'monitoring',
                            'node': 'test_node_1',
                            'pod': 'prometheus-operator-kube-state-metrics-1',
                            'service': 'prometheus-operator-kube-state-metrics'
                        },
                        'values': [
                            [
                                created_at,
                                '1'
                            ],
                            [
                                created_at + lifetime,
                                '1'
                            ]
                        ]
                    },
                    {
                        'metric': {
                            '__name__': 'kube_node_labels',
                            'endpoint': 'http',
                            'instance': '10.48.5.9:8080',
                            'job': 'kube-state-metrics',
                            'label_beta_kubernetes_io_arch': 'amd64',
                            'label_beta_kubernetes_io_fluentd_ds_ready':
                                'true',
                            'label_beta_kubernetes_io_instance_type':
                                'n1-highmem-2',
                            'label_beta_kubernetes_io_os': 'linux',
                            'label_cloud_google_com_gke_nodepool':
                                'logstash-cpu',
                            'label_cloud_google_com_gke_os_distribution':
                                'cos',
                            'label_failure_domain_beta_kubernetes_io_region':
                                'us-west1',
                            'label_failure_domain_beta_kubernetes_io_zone':
                                'us-west1-a',
                            'label_kubernetes_io_hostname': 'test_node_2',
                            'namespace': 'monitoring',
                            'node': 'test_node_2',
                            'pod': 'prometheus-operator-kube-state-metrics-1',
                            'service': 'prometheus-operator-kube-state-metrics'
                        },
                        'values': [
                            [
                                created_at,
                                '1'
                            ],
                            [
                                created_at + lifetime,
                                '1'
                            ]
                        ]
                    },
                ],
                'resultType': 'matrix'
            },
            'status': 'success'
        }


class TestCostGetter(object):

    def fake_requests_get(self, http_request, **_):
        if 'kube_node_created' in http_request:
            return FakeCreationData()
        if 'kube_node_labels' in http_request:
            return FakeLabelData()
        raise ValueError('Your http_request does not contain '
                         'a recognized Grafana metric.')

    @pytest.fixture(autouse=True)
    def monkeypatch(self, monkeypatch):
        monkeypatch.setattr(requests, 'get', self.fake_requests_get)

    def test_init(self):
        # times are intentionally not being cast to ints
        now = time.time()
        nower = now * 2
        # passing start but not end time
        with pytest.raises(ValueError):
            cost.CostGetter(benchmarking_start_time=nower)
        # passing end but not start time
        with pytest.raises(ValueError):
            cost.CostGetter(benchmarking_end_time=0)
        # passing both start and end times
        cost.CostGetter(benchmarking_start_time=now, benchmarking_end_time=now)

    def test_get_time(self):
        cg = cost.CostGetter()  # object creation
        old_time = int(time.time())
        new_time = cg.get_time()
        assert old_time <= new_time

    def test_send_grafana_api_request(self):
        cg = cost.CostGetter()
        response = cg.send_grafana_api_request('kube_node_created')
        assert isinstance(response, dict)

    def test_finish(self):
        start_time = time.time() - 100  # started 100s ago
        cg = cost.CostGetter(benchmarking_start_time=start_time)

        # benchmarking_end_time is not generated until finish() is called
        assert not cg.benchmarking_end_time

        cpu_costs, gpu_costs, total_costs = cg.finish()

        # did benchmarking_end_time get auto-generated?
        assert cg.benchmarking_end_time

        # did entire pipeline process correctly?
        cpu_costs = float(cpu_costs)
        total_costs = float(total_costs)
        gpu_costs = float(gpu_costs)
        assert '%.9f' % (cpu_costs) == '0.000328889'
        assert '%.9f' % (gpu_costs) == '0.002125000'
        assert '%.9f' % (total_costs) == '7.002453889'

    def test_parse_create_response(self):
        # test node exists after benchmarking
        start_time = int(time.time()) - 100  # a little while ago.
        cg = cost.CostGetter(benchmarking_start_time=start_time)

        expected_node_names = ['test_node_1', 'test_node_2']
        lifetime = random.randint(20, 100)

        creation_data = FakeCreationData.json(
            created_at=start_time + 10,  # node started after benchmarking
            lifetime=lifetime)

        node_info = cg.parse_create_response(creation_data)

        for name in expected_node_names:
            assert node_info[name]['lifetime'] == lifetime

        # test node exists before benchmarking
        creation_data = FakeCreationData.json(
            created_at=start_time - 10,  # node started before benchmarking
            lifetime=lifetime)

        node_info = cg.parse_create_response(creation_data)

        for name in expected_node_names:
            assert node_info[name]['lifetime'] == lifetime - 10

    def test_parse_label_response(self):
        # test node exists after benchmarking
        start_time = int(time.time()) - 100  # a little while ago.
        cg = cost.CostGetter(benchmarking_start_time=start_time)

        expected_node_names = ['test_node_1', 'test_node_2']
        expected_preemptibles = [True, False]
        expected_gpus = ['nvidia-tesla-v100', None]

        expected = zip(
            expected_node_names,
            expected_preemptibles,
            expected_gpus
        )

        lifetime = random.randint(11, 100)

        creation_data = FakeLabelData.json(
            created_at=start_time + 10,  # node started after benchmarking
            lifetime=lifetime)

        node_info = cg.parse_label_response(creation_data)

        for name, pre, gpu in expected:
            assert node_info[name]['instance_type'] == 'n1-highmem-2'
            assert node_info[name]['preemptible'] is pre
            assert node_info[name]['gpu'] == gpu

        # test node exists before benchmarking
        creation_data = FakeLabelData.json(
            created_at=start_time - 10,  # node started before benchmarking
            lifetime=lifetime)

        node_info = cg.parse_label_response(creation_data)

        for name, pre, gpu in expected:
            assert node_info[name]['instance_type'] == 'n1-highmem-2'
            assert node_info[name]['preemptible'] is pre
            assert node_info[name]['gpu'] == gpu

    def test_compute_costs(self):
        cg = cost.CostGetter()
        node_info = {
            'node1': {
                'lifetime': 3600,
                'instance_type': 'n1-highmem-2',
                'gpu': 'nvidia-tesla-v100',
                'preemptible': False
            },
            'node2': {
                'lifetime': 3600,
                'instance_type': 'n1-highmem-2',
                'gpu': 'nvidia-tesla-v100',
                'preemptible': False
            }
        }
        cpu_costs, gpu_costs, total_costs = cg.compute_costs(node_info)
        assert cpu_costs == 0
        assert gpu_costs == 5.1968
        assert total_costs == 5.1968

    def test_compute_hourly_cost(self):
        cg = cost.CostGetter()
        node_dict = {
            'instance_type': 'n1-highmem-2',
            'gpu': 'nvidia-tesla-v100',
        }
        for preemptible in (True, False):
            node_dict['preemptible'] = preemptible
            k = 'preemptible' if preemptible else 'ondemand'
            expected = (cost.COST_TABLE[node_dict['instance_type']][k] +
                        cost.GPU_TABLE[node_dict['gpu']][k])
            assert cg.compute_hourly_cost(node_dict) == expected



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for JobManager classes"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import random

from PIL import Image
from twisted.internet import defer

import pytest
import pytest_twisted
import requests

from kiosk_client import manager
from kiosk_client import settings


class Bunch(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


def dummy_ssl_redirect(url, **__):
    return Bunch(url=url.replace('http://', 'https://'))


class TestJobManager(object):

    @pytest.fixture(autouse=True)
    def monkeypatch(self, monkeypatch):
        monkeypatch.setattr(requests, 'get', dummy_ssl_redirect)

    def test_init(self):
        mgr = manager.JobManager(
            job_type='job',
            host='localhost',
            model='m:0',
            data_scale='1',
            data_label='1')
        # test bad model value
        with pytest.raises(Exception):
            mgr = manager.JobManager(
                job_type='job',
                host='localhost',
                model='model_no_version',
                data_scale='1',
                data_label='1')
        with pytest.raises(Exception):
            mgr = manager.JobManager(
                job_type='job',
                host='localhost',
                model='model:nonint_version',
                data_scale='1',
                data_label='1')
        # test bad data_scale value
        with pytest.raises(ValueError):
            mgr = manager.JobManager(
                job_type='job',
                host='localhost',
                model='m:0',
                data_scale='one',
                data_label='1')
        # test bad data_label value
        with pytest.raises(ValueError):
            mgr = manager.JobManager(
                job_type='job',
                host='localhost',
                model='m:0',
                data_scale='1',
                data_label='1.3')

    def test__get_host(self, mocker):
        host = 'example.com'
        mgr = manager.JobManager(job_type='job', host=host)

        assert mgr._get_host(host) == 'https://%s' % host

        host = 'http://example.com'
        assert mgr._get_host(host) == host.replace('http://', 'https://', 1)

        def fail(*_, **__):
            raise ValueError('on purpose')

        mocker.patch('requests.get', fail)
        with pytest.raises(RuntimeError):
            mgr._get_host(host)

    def test_make_job(self):
        mgr = manager.JobManager(
            job_type='job',
            host='localhost',
            model='m:0',
            data_scale='1',
            data_label='0')

        j1 = mgr.make_job('test.png')
        j2 = mgr.make_job('test.png')

        assert j1.data_scale == mgr.data_scale == j2.data_scale
        assert j1.data_label == mgr.data_label == j2.data_label

        assert j1.json() == j2.json()

    def test_get_completed_job_count(self):
        mgr = manager.JobManager(host='localhost', job_type='job')

        j1 = mgr.make_job('test.png')
        j2 = mgr.make_job('test.png')

        mgr.all_jobs = [j1, j2]

        j1.status = 'new'
        j2.status = 'new'
        assert mgr.get_completed_job_count() == 0

        j1.status = 'done'
        j2.status = 'failed'
        assert mgr.get_completed_job_count() == 0

        j1.status = 'done'
        j2.status = 'done'
        assert mgr.get_completed_job_count() == 0

        def fake_restart(delay):
            return None

        j1.failed = True
        j1.restart = fake_restart
        j1.is_expired = True
        assert mgr.get_completed_job_count() == 1

        # test patch, summarized but not expired
        j1.failed = False
        j1.status = 'done'
        j1.created_at = 1
        j1.finished_at = 1
        j1.output_url = 1
        j1.is_expired = False
        j1.expire = lambda: None
        assert mgr.get_completed_job_count() == 0

    def test_summarize(self, tmpdir):

        # pylint: disable=unused-argument
        def fake_upload_file(filepath, hash_filename, prefix):
            return filepath

        def fake_upload_file_bad(filepath, hash_filename, prefix):
            return 1 / 0

        def fake_download_file(url, dest):
            return None

        def fake_download_file_bad(url, dest):
            return 1 / 0

        settings.OUTPUT_DIR = str(tmpdir)

        mgr = manager.JobManager(host='localhost', job_type='job',
                                 upload_results=True,
                                 calculate_cost=True)

        fakejson = lambda: {'output_url': 'example.com/json.txt'}
        mgr.all_jobs = [Bunch(output_url='example.com/a.txt', json=fakejson),
                        Bunch(output_url='example.com/b.txt', json=fakejson)]

        # monkey-patches for testing
        mgr.cost_getter.finish = lambda: (1, 2, 3)
        mgr.upload_file = fake_upload_file
        mgr.summarize()

        # test Exceptions
        mgr.cost_getter.finish = lambda: 0 / 1
        mgr.upload_file = fake_upload_file_bad
        mgr.summarize()

    @pytest_twisted.inlineCallbacks
    def test_check_job_status(self):
        mgr = manager.JobManager(
            host='localhost',
            job_type='job',
            refresh_rate=0)

        mgr.all_jobs = list(range(5))

        global _is_stopped
        _is_stopped = False

        @pytest_twisted.inlineCallbacks
        def dummy_stop():
            global _is_stopped
            _is_stopped = True
            yield defer.returnValue(_is_stopped)

        global _status_counter
        _status_counter = 0

        def get_completed_job_count():
            global _status_counter
            _status_counter += 1
            return _status_counter

        mgr.get_completed_job_count = get_completed_job_count
        mgr._stop = dummy_stop
        mgr.summarize = lambda: True

        _ = yield mgr.check_job_status()
        assert _status_counter == len(mgr.all_jobs)
        assert _is_stopped


class TestBenchmarkingJobManager(object):

    @pytest_twisted.inlineCallbacks
    def test_run(self, tmpdir, mocker):
        tmpdir = str(tmpdir)
        mocker.patch('requests.get', dummy_ssl_redirect)
        mgr = manager.BenchmarkingJobManager(host='localhost', job_type='job')

        # pylint: disable=unused-argument
        def dummy_upload_file(filepath, **kwargs):
            return filepath

        def dummy_start(delay, upload=False):
            return True

        def make_job(*args, **kwargs):
            m = manager.BatchProcessingJobManager(
                host='localhost',
                job_type='job')

            j = m.make_job(*args, **kwargs)
            j.start = dummy_start
            return j

        mgr.check_job_status = lambda: True
        mgr.upload_file = dummy_upload_file
        mgr.make_job = make_job
        mgr.sleep = lambda x: True
        mgr.get_completed_job_count = lambda: True

        valid_image = os.path.join(tmpdir, 'image.png')
        img = Image.new('RGB', (800, 1280), (255, 255, 255))
        img.save(valid_image, 'PNG')

        yield mgr.run(valid_image, count=2, upload=True)

        yield mgr.run(valid_image, count=2, upload=False)


class TestBatchProcessingJobManager(object):

    @pytest_twisted.inlineCallbacks
    def test_run(self, tmpdir, mocker):
        tmpdir = str(tmpdir)
        mocker.patch('requests.get', dummy_ssl_redirect)
        mgr = manager.BatchProcessingJobManager(
            host='localhost',
            job_type='job')

        # pylint: disable=unused-argument
        def dummy_upload_file(filepath, **kwargs):
            return filepath

        def dummy_start(delay, upload=False):
            return True

        def make_job(*args, **kwargs):
            m = manager.BatchProcessingJobManager(
                host='localhost',
                job_type='job')

            j = m.make_job(*args, **kwargs)
            j.start = dummy_start
            return j

        mgr.check_job_status = lambda: True
        mgr.upload_file = dummy_upload_file
        mgr.make_job = make_job

        num = random.randint(1, 5)
        imagename = lambda x: 'image%s.png' % x

        valid_images = []
        for i in range(num):
            valid_image = os.path.join(tmpdir, imagename(i))
            img = Image.new('RGB', (800, 1280), (255, 255, 255))
            img.save(valid_image, 'PNG')
            valid_images.append(valid_image)

        yield mgr.run(tmpdir)



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from os import path
from setuptools import setup
from setuptools import find_packages

# read the contents of your README file
with open(path.join(path.abspath(path.dirname(__file__)), 'README.md')) as f:
    long_description = f.read()


VERSION = '0.7.0'


setup(name='Kiosk_Client',
      version=VERSION,
      description='ClI client for the DeepCell Kiosk.',
      author='Van Valen Lab',
      author_email='vanvalenlab@gmail.com',
      url='https://github.com/vanvalenlab/kiosk-client',
      download_url='https://github.com/vanvalenlab/'
                   'kiosk-client/tarball/{}'.format(VERSION),
      license='LICENSE',
      install_requires=['boto3',
                        'google-cloud-storage',
                        'Pillow',
                        'python-decouple',
                        'python-dateutil',
                        'treq==20.3.0',
                        'twisted[tls]>=20.3.0'],
      extras_require={
          'tests': ['pytest',
                    'pytest-twisted',
                    'pytest-pep8',
                    'pytest-cov',
                    'pytest-mock',
                    'attrs>=19.2.0'],
      },
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8'
      ])



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for Job class"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import os
import random
import timeit

import pytest
import pytest_twisted

from twisted.internet import defer

from kiosk_client import job

global FAILED
FAILED = False  # global toggle for failed responses


class Bunch(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class DummyResponse(object):

    def __init__(self):
        self.phrase = b'PHRASE'
        self.request = Bunch(method=b'POST', absoluteURI=b'localhost')

    @property
    def code(self):
        global FAILED
        if FAILED:
            return 500
        return 200

    @pytest_twisted.inlineCallbacks
    def json(self):
        global FAILED
        if FAILED:
            FAILED = False
            yield defer.returnValue(dict({'success': True}))
        else:
            FAILED = True
            raise AttributeError('on purpose')


def _get_default_job(filepath='filepath.png'):
    return job.Job(
        filepath=filepath,
        host='localhost',
        model_name='model_name',
        model_version='0',
        download_results=True,
        update_interval=0.0001)


class TestJob(object):

    def test_basic(self):
        # create basic job
        j = _get_default_job()

        # properties should be Fals as job has not yet been started
        assert not j.is_done
        assert not j.is_summarized
        # for JSON formatting list values
        j.prediction_time = list(str(x) for x in range(5))
        j.predict_retries = list(str(x) for x in range(5))
        assert isinstance(j.json(), dict)

        # set the status to failed
        j.status = 'failed'
        assert j.is_done
        assert j.is_summarized

        # set the status to done
        j.status = 'done'
        assert j.is_done
        assert not j.is_summarized  # status=done AND other fields are not None

        # model_version should be an integer
        with pytest.raises(ValueError):
            job.Job(filepath='test.png',
                    host='localhost',
                    model_name='model',
                    model_version='version')

        # data_scale should be a float
        with pytest.raises(ValueError):
            job.Job(filepath='test.png',
                    host='localhost',
                    model_name='model',
                    model_version='version',
                    data_scale='one')

        # data_label should be an int
        with pytest.raises(ValueError):
            job.Job(filepath='test.png',
                    host='localhost',
                    model_name='model',
                    model_version='1',
                    data_label='3.14')

    def test__log_http_response(self):
        now = timeit.default_timer()
        j = _get_default_job()

        dummy_response = DummyResponse()
        j._log_http_response(dummy_response, now)
        dummy_response.failed = True
        j._log_http_response(dummy_response, now)

    def test__make_post_request(self):
        j = _get_default_job()
        req = j._make_post_request('localhost', data={})
        assert isinstance(req, defer.Deferred)

    @pytest_twisted.inlineCallbacks
    def test_upload_file(self, tmpdir):

        @pytest_twisted.inlineCallbacks
        def dummy_request_success(*_, **__):
            yield defer.returnValue({'uploadedName': 'uploads/blah.png'})

        @pytest_twisted.inlineCallbacks
        def dummy_request_fail(*_, **__):
            yield defer.returnValue(dict())

        filepath = 'test.png'
        p = tmpdir.join(filepath)
        p.write('content')
        j = _get_default_job(filepath=str(p))

        j._retry_post_request_wrapper = dummy_request_success
        uploaded_path = yield j.upload_file()
        assert uploaded_path == 'uploads/blah.png'

        filepath = 'test2.png'
        p = tmpdir.join(filepath)
        p.write('content')
        j = _get_default_job(filepath=str(p))

        j._retry_post_request_wrapper = dummy_request_fail
        job_id = yield j.upload_file()
        assert job_id is None

    @pytest_twisted.inlineCallbacks
    def test_download_output(self, tmpdir, mocker):

        global _download_failed
        _download_failed = False

        @pytest_twisted.inlineCallbacks
        def send_get_request(_, **__):
            global _download_failed
            if _download_failed:
                _download_failed = False
                response = Bunch(collect=lambda x: x(b'success'))
                yield defer.returnValue(response)
            else:
                _download_failed = True
                errs = _get_default_job()._http_errors
                err = errs[random.randint(0, len(errs) - 1)]
                raise err('on purpose')

        j = _get_default_job()
        j.output_url = 'fakeURL.com/testfile.txt'
        mocker.patch('kiosk_client.job.get_download_path',
                     lambda: str(tmpdir))
        mocker.patch('treq.get', send_get_request)

        result = yield j.download_output()
        assert os.path.isfile(result)
        assert str(result).startswith(str(tmpdir))
        with open(result, 'r') as f:
            assert f.read() == 'success'

    @pytest_twisted.inlineCallbacks
    def test_summarize(self):
        j = _get_default_job()

        j.status = 'failed'
        j.get_redis_value = lambda x: True

        value = yield j.summarize()
        assert value

    @pytest_twisted.inlineCallbacks
    def test_monitor(self):
        j = _get_default_job()

        global _monitor_counter
        _monitor_counter = 0

        @pytest_twisted.inlineCallbacks
        def get_redis_value(_):
            global _monitor_counter
            _monitor_counter += 1

            if _monitor_counter > 3:
                yield defer.returnValue('done')
            else:
                yield defer.returnValue(_monitor_counter)

        j.get_redis_value = get_redis_value
        results = yield j.monitor()
        assert results
        assert results == j.is_done

    @pytest_twisted.inlineCallbacks
    def test_restart(self):

        @pytest_twisted.inlineCallbacks
        def _dummy():
            yield defer.returnValue(True)

        # test no job_id
        j = _get_default_job()
        j.start = _dummy
        result = yield j.restart(0.00001)
        assert result

        # test is_done
        j = _get_default_job()
        j.job_id = 1
        j.status = 'done'
        j.summarize = _dummy
        result = yield j.restart(0.000001)
        assert result

        # test not is_done
        j = _get_default_job()
        j.job_id = 1
        j.status = 'in-progress'
        j.monitor = _dummy
        result = yield j.restart(0.000001)
        assert result

    @pytest_twisted.inlineCallbacks
    def test_create(self):

        @pytest_twisted.inlineCallbacks
        def dummy_request_success(*_, **__):
            yield defer.returnValue({'hash': 'dummy_hash'})

        @pytest_twisted.inlineCallbacks
        def dummy_request_fail(*_, **__):
            yield defer.returnValue(dict())

        j = _get_default_job()
        j._retry_post_request_wrapper = dummy_request_success
        job_id = yield j.create()
        assert job_id == 'dummy_hash'

        j = _get_default_job()
        j._retry_post_request_wrapper = dummy_request_fail
        job_id = yield j.create()
        assert job_id is None

    @pytest_twisted.inlineCallbacks
    def test_get_redis_value(self):

        @pytest_twisted.inlineCallbacks
        def dummy_request_success(*_, **__):
            yield defer.returnValue({'value': 'done'})

        @pytest_twisted.inlineCallbacks
        def dummy_request_fail(*_, **__):
            yield defer.returnValue(dict())

        j = _get_default_job()
        j._retry_post_request_wrapper = dummy_request_success
        job_id = yield j.get_redis_value('status')
        assert job_id == 'done'

        j = _get_default_job()
        j._retry_post_request_wrapper = dummy_request_fail
        job_id = yield j.get_redis_value('status')
        assert job_id is None

    @pytest_twisted.inlineCallbacks
    def test_expire(self):

        @pytest_twisted.inlineCallbacks
        def dummy_request_success(*_, **__):
            yield defer.returnValue({'value': 'done'})

        @pytest_twisted.inlineCallbacks
        def dummy_request_fail(*_, **__):
            yield defer.returnValue(dict())

        j = _get_default_job()
        j._retry_post_request_wrapper = dummy_request_success
        job_id = yield j.expire()
        assert job_id == 'done'

        j = _get_default_job()
        j._retry_post_request_wrapper = dummy_request_fail
        job_id = yield j.expire()
        assert job_id is None

    @pytest_twisted.inlineCallbacks
    def test_start(self):

        @pytest_twisted.inlineCallbacks
        def dummy_request_success(*_, **__):
            yield defer.returnValue(True)

        @pytest_twisted.inlineCallbacks
        def dummy_upload_success(*_, **__):
            yield defer.returnValue('uploads/test.png')

        @pytest_twisted.inlineCallbacks
        def dummy_download_success(*_, **__):
            yield defer.returnValue('downloads/test-results.png')

        @pytest_twisted.inlineCallbacks
        def dummy_request_fail(*_, **__):
            yield defer.returnValue(None)

        # test successful path
        j = _get_default_job()
        j.create = dummy_request_success
        j.summarize = dummy_request_success
        j.monitor = dummy_request_success
        j.get_redis_value = dummy_request_success
        j.expire = dummy_request_success
        j.upload_file = dummy_upload_success
        j.download_output = dummy_download_success

        # is_done and is_summarized
        j.status = 'done'
        j.output_url = 'local'
        j.created_at = datetime.datetime.now().isoformat()
        j.finished_at = datetime.datetime.now().isoformat()

        delay = .000001
        upload = True
        value = yield j.start(delay, upload)
        assert value

        # test status is done but not summarized
        j.output_url = None
        value = yield j.start(delay, upload)
        assert value is False  # raise exception which is caught

        # test status is failed
        j.status = 'failed'
        value = yield j.start(delay, upload)
        assert value

        # test bad results
        j.create = dummy_request_fail
        value = yield j.start(delay, upload)
        assert value is False  # failed

    @pytest_twisted.inlineCallbacks
    def test__retry_post_request_wrapper(self, mocker):

        global _make_request_failed
        _make_request_failed = False

        @pytest_twisted.inlineCallbacks
        def dummy_post_request(*_, **__):
            global _make_request_failed
            if _make_request_failed:
                _make_request_failed = False
                yield defer.returnValue(DummyResponse())
            else:
                _make_request_failed = True
                errs = _get_default_job()._http_errors
                err = errs[random.randint(0, len(errs) - 1)]
                raise err('on purpose')

        j = _get_default_job()
        mocker.patch('treq.post', dummy_post_request)
        result = yield j._retry_post_request_wrapper('host', {})
        assert result.get('success')



# Copyright 2016-2020 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/kiosk-client/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Job class using Twisted to create and monitor itself asynchronously"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os
import timeit

import dateutil.parser
import treq
from twisted.internet import defer
from twisted.internet import error as twisted_errors
from twisted.web import _newclient as twisted_client

from kiosk_client.utils import sleep, strip_bucket_prefix, get_download_path


class Job(object):

    def __init__(self, host, filepath, model_name, model_version, **kwargs):
        """Creates and tracks a DeepCell Kiosk job, recording various summary data.

        Args:
            host (str): public IP address of the DeepCell Kiosk cluster.
            filepath (str): The filepath of the file to be processed.
            model_name (str): Name of servable model.
            model_version (int): Version of servable model.
            kwargs (dict): Optional keyword arguments.
        """
        self.logger = logging.getLogger(str(self.__class__.__name__))

        self.host = str(host)
        self.filepath = str(filepath)
        self.model_name = str(model_name)
        self.model_version = str(model_version)
        self.job_type = str(kwargs.get('job_type', 'segmentation'))

        data_scale = str(kwargs.get('data_scale', ''))
        if data_scale:
            try:
                data_scale = float(data_scale)
            except ValueError:
                raise ValueError('data_scale must be a number.')
        self.data_scale = data_scale

        data_label = str(kwargs.get('data_label', ''))
        if data_label:
            try:
                data_label = int(data_label)
            except ValueError:
                raise ValueError('data_label must be an integer.')
        self.data_label = data_label

        if self.model_version and not self.model_version.isdigit():
            raise ValueError('`model_version` must be a number, got ' +
                             self.model_version)

        self.preprocess = kwargs.get('preprocess', '')
        self.postprocess = kwargs.get('postprocess', '')
        self.upload_prefix = kwargs.get('upload_prefix', 'uploads')
        self.upload_prefix = strip_bucket_prefix(self.upload_prefix)
        self.expire_time = int(kwargs.get('expire_time', 3600))
        self.update_interval = int(kwargs.get('update_interval', 10))
        self.original_name = kwargs.get('original_name', self.filepath)
        self.download_results = kwargs.get('download_results', False)

        self.failed = False  # for error handling
        self.is_expired = False

        self.headers = {
            'Content-Type': ['application/json'],
            'Connection': 'close',
        }

        # summary data
        self.status = None
        self.job_id = None
        self.created_at = None
        self.finished_at = None
        self.postprocess_time = None
        self.prediction_time = None
        self.download_time = None
        self.upload_time = None
        self.output_url = None
        self.total_jobs = None
        self.total_time = None
        self.reason = None
        self.children_upload_time = None
        self.cleanup_time = None
        self.predict_retries = None
        self._finished_statuses = {'done', 'failed'}

        self.pool = kwargs.get('pool')

        self.sleep = sleep  # allow monkey-patch

        self._http_errors = (
            twisted_client.ResponseNeverReceived,
            twisted_client.RequestTransmissionFailed,
            twisted_errors.ConnectBindError,
            twisted_errors.TimeoutError,
            twisted_errors.ConnectError,
            twisted_errors.ConnectionRefusedError,
        )

    @property
    def is_done(self):
        return self.status in self._finished_statuses

    @property
    def is_summarized(self):
        if self.status == 'failed':
            return True
        summaries = (self.created_at, self.finished_at, self.output_url)
        is_summarized = all(x is not None for x in summaries)
        return is_summarized and self.is_done

    def json(self):

        def _float(x):
            if isinstance(x, list):
                return [_float(y) for y in x]
            try:
                return float(x)
            except:  # pylint: disable=bare-except
                return x

        return {
            'input_file': self.original_name,
            'status': self.status,
            'total_time': _float(self.total_time),
            'total_jobs': _float(self.total_jobs),
            'download_url': self.output_url,
            'created_at': self.created_at,
            'finished_at': self.finished_at,
            'prediction_time': _float(self.prediction_time),
            'postprocess_time': _float(self.postprocess_time),
            'upload_time': _float(self.upload_time),
            'download_time': _float(self.download_time),
            'predict_retries': _float(self.predict_retries),
            'cleanup_time': _float(self.cleanup_time),
            'children_upload_time': _float(self.children_upload_time),
            'model': '{}:{}'.format(self.model_name, self.model_version),
            'postprocess': self.postprocess,
            'preprocess': self.preprocess,
            'reason': self.reason,
            'job_id': self.job_id,
        }

    def _log_http_response(self, response, created_at):
        log = self.logger.debug if response.code == 200 else self.logger.warning
        log('%s %s - %s %s - took %ss',
            response.request.method.decode(),
            response.request.absoluteURI.decode(),
            response.code, response.phrase.decode(),
            timeit.default_timer() - created_at)

    def _make_post_request(self, host, **kwargs):
        req_kwargs = {
            'headers': kwargs.get('headers', self.headers),
            'pool': kwargs.get('pool', self.pool),
        }

        # The name of the payload dictates the type of encoding and headers.
        payload_names = {'data', 'json', 'files'}
        for pn in payload_names:
            if pn in kwargs:
                req_kwargs[pn] = kwargs[pn]

        return treq.post(host, **req_kwargs)

    @defer.inlineCallbacks
    def _retry_post_request_wrapper(self, host, name='REDIS', **kwargs):
        retrying = True  # retry loop to prevent stackoverflow
        while retrying:
            created_at = timeit.default_timer()
            try:
                request = self._make_post_request(host, **kwargs)
                response = yield request  # Wait for the deferred request
            except self._http_errors as err:
                self.logger.warning('[%s]: Encountered %s during %s: %s',
                                    self.job_id, type(err).__name__, name, err)
                yield self.sleep(self.update_interval)
                continue  # return to top of retry loop

            try:
                self._log_http_response(response, created_at)
                json_content = yield response.json()  # parse the JSON data
            except (ValueError, AttributeError) as err:
                self.logger.error('[%s]: Failed to parse %s response as JSON '
                                  'due to %s: %s', self.job_id, name,
                                  type(err).__name__, err)
                yield self.sleep(self.update_interval)
                continue  # return to top of retry loop

            retrying = False  # success

        defer.returnValue(json_content)  # "return" the value

    @defer.inlineCallbacks
    def upload_file(self):
        host = '{}/api/upload'.format(self.host)
        headers = self.headers.copy()
        headers['Content-Type'] = ['multipart/form-data']
        name = 'UPLOAD {}'.format(self.filepath)
        payload = {
            'file': (self.filepath, open(self.filepath, 'rb'))
        }
        response = yield self._retry_post_request_wrapper(host, name,
                                                          files=payload,
                                                          headers=headers)
        uploaded_path = response.get('uploadedName')
        defer.returnValue(uploaded_path)  # "return" the value

    @defer.inlineCallbacks
    def get_redis_value(self, field):
        host = '{}/api/redis'.format(self.host)
        payload = {'hash': self.job_id, 'key': field}
        name = 'REDIS HGET {}'.format(field)
        response = yield self._retry_post_request_wrapper(host, name,
                                                          json=payload)
        value = response.get('value')
        defer.returnValue(value)  # "return" the value

    @defer.inlineCallbacks
    def create(self):
        # Build a deferred request to the create API
        # See https://tinyurl.com/u3qs9om for expected POST data
        job_data = {
            'modelName': self.model_name,
            'modelVersion': self.model_version,
            'preprocessFunction': self.preprocess,
            'postprocessFunction': self.postprocess,
            'imageName': self.filepath,
            'jobType': self.job_type,
            'dataRescale': self.data_scale,
            'dataLabel': self.data_label,
            'uploadedName': os.path.join(self.upload_prefix, self.filepath),
        }
        host = '{}/api/predict'.format(self.host)
        name = 'REDIS CREATE'
        response = yield self._retry_post_request_wrapper(host, name,
                                                          json=job_data)

        job_id = response.get('hash')

        if job_id is not None:
            self.logger.debug('[%s]: Successfully created.', job_id)
        else:
            self.logger.error('Create response JSON is invalid: %s', response)

        defer.returnValue(job_id)  # "return" the value

    @defer.inlineCallbacks
    def monitor(self):
        while not self.is_done:

            yield self.sleep(self.update_interval)  # prevent 429s

            status = yield self.get_redis_value('status')

            if self.status != status:
                self.status = status
                self.logger.info('[%s]: Found new %sstatus `%s`.', self.job_id,
                                 'final ' if self.is_done else '', self.status)

        defer.returnValue(self.is_done)  # "return" the value

    @defer.inlineCallbacks
    def summarize(self):
        summary_attributes = (
            'created_at',
            'finished_at',
            'reason',
            'output_url',
        )
        attributes = (
            'prediction_time',
            'predict_retries',
            'postprocess_time',
            'upload_time',
            'download_time',
            'children_upload_time',
            'cleanup_time',
            'total_jobs',
            'total_time',
        )
        # get the string values
        for name in summary_attributes:
            value = yield self.get_redis_value(name)
            setattr(self, name, value)  # save the valid value to self

        # get the numerical values and parse into list if required
        for name in attributes:
            value = yield self.get_redis_value(name)
            value = str(value).split(',')
            if len(value) == 1:
                value = value[0]
            setattr(self, name, value)  # save the valid value to self

        defer.returnValue(self.is_summarized)  # "return" the value

    @defer.inlineCallbacks
    def expire(self):
        host = '{}/api/redis/expire'.format(self.host)
        payload = {'hash': self.job_id, 'expireIn': self.expire_time}
        name = 'REDIS EXPIRE'
        response = yield self._retry_post_request_wrapper(host, name,
                                                          json=payload)
        value = response.get('value')
        defer.returnValue(value)  # "return" the value

    @defer.inlineCallbacks
    def download_output(self):
        start = timeit.default_timer()
        basename = self.output_url.split('/')[-1]
        dest = os.path.join(get_download_path(), basename)
        self.logger.info('[%s]: Downloading output file %s to %s.',
                         self.job_id, self.output_url, dest)
        name = 'DOWNLOAD RESULTS'
        retrying = True  # retry loop to prevent stackoverflow
        while retrying:
            try:
                request = treq.get(self.output_url, unbuffered=True)
                response = yield request
            except self._http_errors as err:
                self.logger.warning('[%s]: Encountered %s during %s: %s',
                                    self.job_id, type(err).__name__, name, err)
                yield self.sleep(self.update_interval)
                continue  # return to top of retry loop
            retrying = False  # success

        with open(dest, 'wb') as outfile:
            yield response.collect(outfile.write)

        self.logger.info('Saved output file: "%s" in %s s.',
                         dest, timeit.default_timer() - start)

        defer.returnValue(dest)

    @defer.inlineCallbacks
    def restart(self, delay=0):
        if not self.failed:
            self.logger.warning('[%s]: Restarting but not failed.', self.job_id)

        self.failed = False  # reset failure mode to prevent further restarts

        if delay:
            yield self.sleep(delay)

        self.logger.debug('[%s]: Restarting failed job.', self.job_id)

        if self.job_id is None:  # never got started in the first place
            result = yield self.start()

        elif self.is_done:  # no need to monitor, skip straight to summarize
            result = yield self.summarize()

        else:  # job has begun but was not finished, monitor status
            result = yield self.monitor()

        defer.returnValue(result)

    @defer.inlineCallbacks
    def start(self, delay=0, upload=False):

        if delay:  # delay the start if required
            yield self.sleep(delay)

        if upload:
            uploaded_path = yield self.upload_file()
            self.filepath = os.path.relpath(uploaded_path, self.upload_prefix)

        try:
            self.job_id = yield self.create()
            assert self.job_id is not None, 'Create did not return a job ID'

            success = yield self.monitor()
            assert success, 'Monitor did not have a successful return vaue'

            success = yield self.summarize()
            assert success, 'Summarize did not have a successful return vaue'

            if self.status == 'done' and self.is_summarized:
                # TODO: `dateutil` deprecated by python 3.7 `fromisoformat`
                # created_at = datetime.datetime.fromisoformat(created_at)
                # finished_at = datetime.datetime.fromisoformat(finished_at)
                created_at = dateutil.parser.parse(self.created_at)
                finished_at = dateutil.parser.parse(self.finished_at)
                diff = finished_at - created_at
                self.logger.info('[%s]: Finished in %s seconds with status '
                                 '`%s`. Download at `%s`.',
                                 self.job_id, diff.total_seconds(),
                                 self.status, self.output_url)

                if self.download_results:
                    success = yield self.download_output()

            elif self.status == 'failed':
                reason = yield self.get_redis_value('reason')
                self.logger.warning('[%s]: Found final status `%s`: %s',
                                    self.job_id, self.status, reason)

            else:
                raise ValueError('Job %s was about to expire with status %s' %
                                 (self.job_id, self.status))

            yield self.sleep(self.update_interval)
            value = yield self.expire()

            assert value == 1, 'Failed to expire key %s' % self.job_id
            self.is_expired = True

            defer.returnValue(value)

        except Exception as err:
            self.failed = True
            self.logger.error('[%s]: Encountered unexpected error in '
                              'job.start(): %s', self.job_id, err)
            defer.returnValue(False)
